from StringIO import StringIO
import os
import sys

from filesystems import Path
from filesystems.exceptions import FileExists, FileNotFound
import click.testing
import filesystems.memory

from mkenv.common import Locator, VirtualEnv


class CLIMixin(object):
    def setUp(self):
        super(CLIMixin, self).setUp()

        self.stdin = StringIO()
        self.stdout = StringIO()
        self.stderr = StringIO()

        self.filesystem = filesystems.memory.FS()

        self.root_dir = Path("virtualenvs")
        self.filesystem.create_directory(self.root_dir)

        self.link_dir = Path("bin")
        self.filesystem.create_directory(self.link_dir)

        self.locator = Locator(
            root=self.root_dir,
            make_virtualenv=lambda **kwargs: VirtualEnv(
                create=self.fake_create,
                install=self.fake_install,
                **kwargs
            ),
        )

    def installed(self, virtualenv):
        base = virtualenv.path
        try:
            with self.filesystem.open(base.descendant("packages")) as f:
                packages = set(line.strip() for line in f)
        except FileNotFound:
            packages = set()

        try:
            with self.filesystem.open(base.descendant("reqs")) as f:
                reqs = set(line.strip() for line in f)
        except FileNotFound:
            reqs = set()

        return packages, reqs

    def fake_create(self, virtualenv, **kwargs):
        try:
            self.filesystem.create_directory(path=virtualenv.path)
        except FileExists:
            pass

    def fake_install(self, virtualenv, packages, requirements, **kwargs):
        base = virtualenv.path
        with self.filesystem.open(base.descendant("packages"), "a") as f:
            f.writelines(
                package.encode("utf-8") + "\n" for package in packages
            )
        with self.filesystem.open(base.descendant("reqs"), "a") as f:
            f.writelines(req.encode("utf-8") + "\n" for req in requirements)

    def run_cli(self, argv=(), exit_status=os.EX_OK):
        runner = click.testing.CliRunner()
        result = runner.invoke(
            self._fix_click(self.cli.main),
            args=argv,
            default_map=dict(
                link_dir=self.link_dir,
                locator=self.locator,
                filesystem=self.filesystem,
            ),
        )
        if result.exception and not isinstance(result.exception, SystemExit):
            cls, exc, tb = result.exc_info
            raise cls, exc, tb

        self.assertEqual(
            result.exit_code,
            exit_status,
            msg="Different exit code, {} != {}\n\nstderr:\n\n{!r}".format(
                result.exit_code, exit_status, self.stderr.getvalue(),
            ),
        )
        return (
            self.stdin.getvalue(),
            self.stdout.getvalue(),
            self.stderr.getvalue(),
        )

    def _fix_click(self, real_main):
        """
        Click is really really really annoying.

        It patches sys.stdout and sys.stderr to the same exact StringIO.

        """

        class Fixed(object):
            def __getattr__(self, attr):
                return getattr(real_main, attr)

            def main(this, *args, **kwargs):
                stdout, sys.stdout = sys.stdout, self.stdout
                stderr, sys.stderr = sys.stderr, self.stderr
                real_main.main(*args, **kwargs)
                sys.stdout = stdout
                sys.stderr = stderr
        return Fixed()



from unittest import TestCase
import os

from mkenv import remove
from mkenv.tests.utils import CLIMixin


class TestRemove(CLIMixin, TestCase):

    cli = remove

    def test_remove_removes_an_env_with_the_given_name(self):
        boom = self.locator.for_name("boom")
        boom.create()
        self.assertTrue(boom.exists_on(filesystem=self.filesystem))
        self.run_cli(["boom"])
        self.assertFalse(boom.exists_on(filesystem=self.filesystem))

    def test_remove_multiple(self):
        names = ["boom", "bang", "whiz"]
        venvs = [self.locator.for_name(name=name) for name in names]

        for venv in venvs:
            venv.create()

        self.run_cli(names)
        self.assertEqual(
            [venv.exists_on(filesystem=self.filesystem) for venv in venvs],
            [False, False, False],
        )

    def test_cannot_remove_non_existing_envs(self):
        boom = self.locator.for_name("boom")
        self.assertFalse(boom.exists_on(filesystem=self.filesystem))
        self.run_cli(["boom"], exit_status=os.EX_NOINPUT)

    def test_can_remove_non_existing_envs_with_force(self):
        boom = self.locator.for_name("boom")
        self.assertFalse(boom.exists_on(filesystem=self.filesystem))
        self.run_cli(["--force", "boom"])



from unittest import TestCase
import os

from mkenv import converge
from mkenv.tests.utils import CLIMixin


class TestConverge(CLIMixin, TestCase):

    cli = converge

    def test_it_creates_missing_virtualenvs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("c").exists_on(self.filesystem))

        with self.filesystem.open(
            self.locator.root.descendant("virtualenvs.toml"), "w",
        ) as venvs:
            venvs.write(
                """
                [virtualenv.a]
                [virtualenv.b]
                install = ["foo", "bar", "bla"]
                requirements = ["requirements.txt"]
                [virtualenv.c]
                install = ["foo", "$HOME", "~/a"]
                link = ["bar", "baz"]
                """
            )

        self.run_cli([])

        self.assertEqual(
            (
                self.installed(self.locator.for_name("a")),
                self.installed(self.locator.for_name("b")),
                self.installed(self.locator.for_name("c")),
            ), (
                (set(), set()),
                ({"foo", "bar", "bla"}, {"requirements.txt"}),
                ({"foo", os.getenv("HOME"), os.path.expanduser("~/a")}, set()),
            ),
        )

    def test_it_converges_existing_virtualenvs(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("b").exists_on(self.filesystem))
        self.assertFalse(self.locator.for_name("c").exists_on(self.filesystem))

        with self.filesystem.open(
            self.locator.root.descendant("virtualenvs.toml"), "w",
        ) as venvs:
            venvs.write(
                """
                [virtualenv.a]
                install = ["foo", "bar"]
                requirements = ["requirements.txt"]
                """
            )

        self.run_cli([])

        with self.filesystem.open(
            self.locator.root.descendant("virtualenvs.toml"), "w",
        ) as venvs:
            venvs.write(
                """
                [virtualenv.a]
                install = ["baz", "quux"]
                requirements = ["requirements.txt", "other.txt"]
                """
            )

        self.run_cli([])

        self.assertEqual(
            self.installed(self.locator.for_name("a")),
            ({"baz", "quux"}, {"requirements.txt", "other.txt"}),
        )



"""
Converge the set of installed virtualenvs.

"""
import os
import sys

from filesystems.exceptions import FileExists, FileNotFound
import click
import toml

from mkenv.common import _FILESYSTEM, _LINK_DIR, _ROOT


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@_FILESYSTEM
@_LINK_DIR
@_ROOT
def main(filesystem, locator, link_dir):
    with filesystem.open(locator.root.descendant("virtualenvs.toml")) as venvs:
        contents = toml.load(venvs)

    for name, config in contents["virtualenv"].iteritems():
        config.setdefault("python", sys.version)

        virtualenv = locator.for_name(name=name)
        existing_config_path = virtualenv.path.descendant("installed.toml")

        try:
            with filesystem.open(existing_config_path) as existing_config:
                if toml.loads(existing_config.read()) == config:
                    continue
        except FileNotFound:
            virtualenv.create()
        else:
            virtualenv.recreate_on(filesystem=filesystem)

        packages = [
            os.path.expandvars(os.path.expanduser(package))
            for package in config.get("install", [])
        ]
        requirements = [
            os.path.expandvars(os.path.expanduser(requirement))
            for requirement in config.get("requirements", [])
        ]
        virtualenv.install(packages=packages, requirements=requirements)
        for link in config.get("link", []):
            source = virtualenv.binary(name=link)
            try:
                filesystem.link(
                    source=source, to=link_dir.descendant(link),
                )
            except FileExists as error:
                if filesystem.realpath(error.value) != source:
                    raise

        with filesystem.open(existing_config_path, "w") as existing_config:
            existing_config.write(toml.dumps(config).encode("utf-8"))



from unittest import TestCase
import sys

from filesystems import Path
import filesystems.native

from mkenv import find, make
from mkenv.common import Locator
from mkenv.tests.utils import CLIMixin


class TestMake(CLIMixin, TestCase):

    cli = make

    def test_make_creates_an_env_with_the_given_name(self):
        self.assertFalse(self.locator.for_name("a").exists_on(self.filesystem))
        self.run_cli(["a"])
        self.assertTrue(self.locator.for_name("a").exists_on(self.filesystem))

    def test_make_t_creates_a_global_temporary_environment(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists_on(self.filesystem))

        stdin, stdout, stderr = self.run_cli(["--temporary"])
        self.assertEqual(
            (temporary.exists_on(self.filesystem), stdin, stdout, stderr),
            (True, "", str(temporary.path.descendant("bin")) + "\n", ""),
        )

    def test_make_t_recreates_the_environment_if_it_exists(self):
        temporary = self.locator.temporary()
        self.assertFalse(temporary.exists_on(self.filesystem))
        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists_on(self.filesystem))

        foo = temporary.path.descendant("foo")
        self.filesystem.touch(path=foo)
        self.assertTrue(self.filesystem.exists(path=foo))

        self.run_cli(["--temporary"])
        self.assertTrue(temporary.exists_on(self.filesystem))
        self.assertFalse(self.filesystem.exists(path=foo))

    def test_cannot_specify_both_name_and_temporary(self):
        stdin, stdout, stderr = self.run_cli(
            ["--temporary", "foo"], exit_status=2,
        )
        self.assertTrue(
            stderr.endswith(
                "specify only one of '-t / --temp / --temporary' or 'name'\n"
            ), msg=stderr,
        )

    def test_recreate(self):
        virtualenv = self.locator.for_name("something")
        self.assertFalse(virtualenv.exists_on(self.filesystem))

        virtualenv.create()

        thing = virtualenv.path.descendant("thing")
        self.filesystem.touch(path=thing)
        self.assertTrue(self.filesystem.exists(thing))

        self.run_cli(["--recreate", "something"])
        self.assertTrue(virtualenv.exists_on(self.filesystem))

        self.assertFalse(self.filesystem.exists(thing))

    def test_install_and_requirements(self):
        self.run_cli(["-i", "foo", "-i", "bar", "-r", "reqs.txt", "bla"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("bla")),
            ({"foo", "bar"}, {"reqs.txt"}),
        )

    def test_mkenv_default_name(self):
        """
        Just saying ``mkenv`` creates an environment based on the current
        directory's name.

        """

        virtualenv = self.locator.for_directory(Path.cwd())
        self.assertFalse(virtualenv.exists_on(self.filesystem))
        self.run_cli([])
        self.assertTrue(virtualenv.exists_on(self.filesystem))

    def test_install_default_name(self):
        """
        If you install one single package and don't specify a name, the name of
        the installed package is used.

        """

        self.run_cli(["-i", "foo"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("foo")), ({"foo"}, set()),
        )

    def test_install_default_name_with_version_specification(self):
        self.run_cli(["-i", "thing[foo]>=2,<3"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("thing")),
            ({"thing[foo]>=2,<3"}, set()),
        )

    def test_temporary_env_with_single_install(self):
        self.run_cli(["-t", "-i", "thing"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.temporary()), ({"thing"}, set()),
        )

    def test_link_default_name(self):
        """
        If you link one single binary and don't specify a name, the name of
        the binary is probably both the package and what you want to call the
        environment.

        """

        self.run_cli(["-l", "foo"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("foo")), ({"foo"}, set()),
        )

    def test_multiple_installs_one_link(self):
        self.run_cli(["-i", "foo", "-i", "bar", "-l", "foo", "baz"])
        # We've stubbed out our Locator's venvs' install to just store.
        self.assertEqual(
            self.installed(self.locator.for_name("baz")),
            ({"foo", "bar"}, set()),
        )

    def test_multiple_installs_one_link_no_name(self):
        self.run_cli(["-i", "foo", "-i", "bar", "-l", "foo"], exit_status=2)


class TestIntegration(TestCase):
    def setUp(self):
        self.fs = filesystems.native.FS()
        self.root = self.fs.temporary_directory()
        self.addCleanup(self.fs.remove, self.root)

        # Fucking click.
        stdout = sys.stdout
        self.addCleanup(lambda: setattr(sys, "stdout", stdout))

    def test_it_works(self):
        with self.fs.open(self.root.descendant("make_stdout"), "wb") as stdout:
            sys.stdout = stdout

            try:
                make.main(
                    args=[
                        "--root", str(self.root),
                        "mkenv-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        with self.fs.open(self.root.descendant("find_stdout"), "wb") as stdout:
            sys.stdout = stdout

            try:
                find.main(
                    [
                        "--root", str(self.root),
                        "--existing-only",
                        "name", "mkenv-unittest-should-be-deleted",
                    ],
                )
            except SystemExit:
                pass

        locator = Locator(root=self.root)
        virtualenv = locator.for_name("mkenv-unittest-should-be-deleted")
        self.assertEqual(
            self.fs.contents_of(self.root.descendant("find_stdout")),
            str(virtualenv.path) + "\n",
        )



"""
mkenv creates virtualenvs.

By default it places them in the appropriate data directory for your platform
(See `appdirs <https://pypi.python.org/pypi/appdirs>`_), but it will also
respect the :envvar:`WORKON_HOME` environment variable for compatibility with
:command:`mkvirtualenv`.

"""

from functools import partial

from filesystems import Path
from packaging.requirements import Requirement
import click

from mkenv.common import _FILESYSTEM, _LINK_DIR, _ROOT


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@_FILESYSTEM
@_LINK_DIR
@_ROOT
@click.option(
    "-i", "--install", "installs",
    multiple=True,
    help=(
        "install the given specifier (package) into the "
        "virtualenv with pip after it is created"
    ),
)
@click.option(
    "-l", "--link", "links",
    multiple=True,
    help=(
        "After installing any specified packages, link the specified "
        "binaries into the directory they would have been installed into "
        "globally."
    ),
)
@click.option(
    "-r", "--requirement", "requirements",
    multiple=True,
    help=(
        "install the given requirements file into the "
        "virtualenv with pip after it is created"
    ),
)
@click.option(
    "-R", "--recreate",
    flag_value=True,
    help="recreate the virtualenv if it already exists",
)
@click.option(
    "-t", "--temp", "--temporary",
    flag_value=True,
    help="create or reuse the global temporary virtualenv",
)
@click.argument("name", required=False)
@click.argument("virtualenv_args", nargs=-1, type=click.UNPROCESSED)
def main(
    filesystem,
    link_dir,
    name,
    locator,
    temporary,
    installs,
    links,
    requirements,
    recreate,
    virtualenv_args,
):
    if name:
        if temporary:
            raise click.BadParameter(
                "specify only one of '-t / --temp / --temporary' or 'name'",
            )

        virtualenv = locator.for_name(name=name)
    elif temporary:
        virtualenv = locator.temporary()
        click.echo(virtualenv.binary("python").dirname())
        act = partial(virtualenv.recreate_on, filesystem=filesystem)
    elif len(installs) == 1:
        # When there's just one package to install, default to using that name.
        requirement, = installs
        name = Requirement(requirement).name
        virtualenv = locator.for_name(name=name)
    elif installs:
        raise click.BadParameter("A name is required.")
    elif len(links) == 1:
        # When there's just one binary to link, go for the gold.
        name, = installs = links
        virtualenv = locator.for_name(name=name)
    else:
        virtualenv = locator.for_directory(directory=Path.cwd())

    if recreate or temporary:
        act = partial(virtualenv.recreate_on, filesystem=filesystem)
    else:
        act = virtualenv.create

    act(arguments=virtualenv_args)
    virtualenv.install(packages=installs, requirements=requirements)

    for link in links:
        filesystem.link(
            source=virtualenv.binary(name=link),
            to=link_dir.descendant(link),
        )



from unittest import TestCase

from filesystems import Path

from mkenv.common import Locator, VirtualEnv


class TestLocator(TestCase):
    def test_named_virtualenvs_are_children(self):
        locator = Locator(root=Path.root())
        self.assertEqual(
            locator.for_name("one"),
            VirtualEnv(path=locator.root.descendant("one")),
        )



from functools import partial

from filesystems import Path
import click
import sys

from mkenv.common import _FILESYSTEM, _ROOT, PATH


def run(
    locator,
    filesystem,
    binary=None,
    directory=None,
    name=None,
    existing_only=False,
):
    """
    Find the virtualenv associated with a given project given its name or path.

    If an optional binary is provided, the binary's path within the virtualenv
    is returned.

    """

    if directory is not None:
        virtualenv = locator.for_directory(directory=directory)
    else:
        if name is None:
            sys.stdout.write(str(locator.root))
            sys.stdout.write("\n")
            return

        virtualenv = locator.for_name(name=name)

    if existing_only and not virtualenv.exists_on(filesystem=filesystem):
        return 1

    if binary is not None:
        sys.stdout.write(str(virtualenv.binary(binary)))
    else:
        sys.stdout.write(str(virtualenv.path))

    sys.stdout.write("\n")


@click.group(
    context_settings=dict(help_option_names=["-h", "--help"]),
    invoke_without_command=True,
)
@_FILESYSTEM
@_ROOT
@click.option(
    "-E", "--existing-only",
    flag_value=True,
    help="Only consider existing virtualenvs.",
)
@click.pass_context
def main(context, locator, existing_only, filesystem):
    if context.invoked_subcommand is None:
        click.echo(locator.root)
    else:
        context.obj = dict(
            locate=partial(
                run,
                locator=locator,
                existing_only=existing_only,
                filesystem=filesystem,
            ),
        )


@main.command()
@click.argument("directory", required=False, type=PATH)
@click.argument("binary", required=False)
@click.pass_context
def directory(context, directory, binary):
    """
    Find the virtualenv given the project's path.

    """

    locate = context.obj["locate"]
    context.exit(
        locate(directory=directory or Path.cwd(), binary=binary) or 0,
    )


@main.command()
@click.argument("name")
@click.argument("binary", required=False)
@click.pass_context
def name(context, name, binary):
    """
    Find the virtualenv given the project's name.

    """

    locate = context.obj["locate"]
    context.exit(locate(name=name, binary=binary) or 0)
